"""Demonstrates basic usage of class-based views with FastAPI."""
import typing

import fastapi
import pydantic

import cbra


class ExampleResourceModel(pydantic.BaseModel):
    object_id: int


class ExampleSubresourceModel(pydantic.BaseModel):
    object_id: int
    child_id: int


class Subresource(cbra.PublicResource):
    path_parameter: int = 'child_id'
    name: str = 'sub'

    class Meta:
        name: str = 'subresource'
        path_parameter: int = 'child_id'

    async def list(self, object_id: str):
        return {'object_id': object_id}

    async def retrieve(self, object_id: str, child_id: int):
        return {'object_id': object_id, 'child_id': child_id}


class ExampleResource(cbra.PublicResource):
    name: str = 'example'
    path_parameter: str = 'object_id'
    subresources = [Subresource]
    resource_class = ExampleSubresourceModel

    class Meta:
        name: str = 'example'
        path_parameter: str = 'object_id'

    async def create(self, dto: ExampleSubresourceModel) -> typing.Union[ExampleResourceModel, ExampleSubresourceModel]:
        return dto

    async def list(self):
        return []

    async def destroy(self, object_id: str):
        return {'object_id': object_id}

    async def exists(self, object_id: str):
        return {'object_id': object_id}

    async def receive(self, object_id: str, dto: dict):
        return {'object_id': object_id, 'dto': dto}

    async def replace(self, object_id: str, dto: dict):
        return {'object_id': object_id, 'dto': dto}

    async def retrieve(self, object_id: str):
        return {'object_id': object_id}

    async def update(self, object_id: str, dto: dict):
        return {'object_id': object_id, 'dto': dto}


class ExampleProtectedResource(cbra.Resource):
    name: str = 'protected'
    path_parameter: str = 'object_id'

    class Meta:
        name: str = 'secret'
        path_parameter: str = 'object_id'

    async def retrieve(self, object_id: str):
        return {'object_id': object_id, 'principal': self.principal}


app = cbra.Application(
    docs_url='/ui',
    redoc_url='/docs'
)
ExampleResource.add_to_router(app)
ExampleProtectedResource.add_to_router(app)


if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)



import cbra
import uvicorn
from cbra.ext import ioc


class Foo:

    async def on_asgi_boot(self, app: cbra.Application):
        print("Running boot function")

    async def on_asgi_shutdown(self, app: cbra.Application):
        print("Running shutdown function")


ioc.provide('Foo', Foo())
app = cbra.Application()

if __name__ == '__main__':
    uvicorn.run('__main__:app', port=5000)



"""Declares :class:`GrantType`."""
import enum


__all__ = [
    'GrantType',
    'AssertedGrantType'
]


class GrantType(str, enum.Enum):
    __module__: str = 'cbra.ext.oauth2.types'
    authorization_code = "authorization_code"
    client_credentials = "client_credentials"
    refresh_token = "refresh_token"
    jwt_bearer = "urn:ietf:params:oauth:grant-type:jwt-bearer"
    saml_bearer = "urn:ietf:params:oauth:grant-type:saml2-bearer"
    session = "urn:webid:params:oauth:grant-type:session"


class AssertedGrantType(str, enum.Enum):
    __module__: str = 'cbra.ext.oauth2.types'
    jwt_bearer = "urn:ietf:params:oauth:grant-type:jwt-bearer"
    saml_bearer = "urn:ietf:params:oauth:grant-type:saml2-bearer"


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`Event`."""
import datetime

import pydantic

from .resourcelinks import ResourceLinks


class Event(pydantic.BaseModel):
    __module__: str = 'cbra.ext.woocommerce.models'

    id: int = pydantic.Field(
        default=...,
        title="ID",
        description=(
            "Identifies the resource on which an event occurred, in combination "
            "with the specified resource type in the `X-WC-Webhook-Resource` "
            "request header."
        )
    )

    date_created: datetime.datetime = pydantic.Field(
        default=...,
        title="Created",
        description=(
            "The date and time at which the object was created."
        )
    )

    date_created_gmt: datetime.datetime = pydantic.Field(
        default=...,
        title="Created (UTC)",
        description=(
            "The date and time at which the object was created (UTC)."
        )
    )

    date_modified: datetime.datetime = pydantic.Field(
        default=...,
        title="Modified",
        description=(
            "The date and time at which the object was last modified."
        )
    )

    date_modified_gmt: datetime.datetime = pydantic.Field(
        default=...,
        title="Modified (UTC)",
        description=(
            "The date and time at which the object was last modified (UTC)."
        )
    )

    links: ResourceLinks = pydantic.Field(
        default=...,
        alias="_links",
        title="Links"
    )

    def get_resource_uri(self) -> str:
        return self.links.self[0].href


# pylint: skip-file
from . import const
from .basesession import BaseSession
from .cookie import CookieSession
from .null import NullSession
from .sessionrequired import SessionRequired


__all__: list[str] = [
    'const',
    'BaseSession',
    'CookieSession',
    'NullSession',
    'SessionRequired',
]


"""Declares :class:`DependencyList`."""
import asyncio
import typing

import pydantic

from .dependency import Dependency


class DependencyList(pydantic.BaseModel):
    """A list of dependencies that can be resolved by the provider."""
    __module__: str = 'cbra.ext.ioc.models'
    force: bool = True
    items: typing.List[Dependency]

    @staticmethod
    async def _resolve(dependency: Dependency) -> typing.Tuple[Dependency, typing.Any]:
        return (dependency, await dependency.resolve())

    async def resolve(self) -> list[tuple[Dependency, typing.Any]]:
        """Resolve all dependencies in the list."""
        futures = [self._resolve(x) for x in self.items]
        return [x for x in await asyncio.gather(*futures)]



"""Declares :class:`QueryOptions`"""
import inspect
import typing

import fastapi
import pydantic


class QueryOptions(pydantic.BaseModel):
    offset: typing.Optional[int] = 0
    limit: typing.Optional[int] = 100
    order_by: typing.Optional[typing.List[str]] = []

    @classmethod
    def is_reserved_field(cls, name: str) -> bool:
        return name in {'offset', 'limit', 'order_by'}

    @classmethod
    def null(cls) -> typing.Optional['QueryOptions']:
        return None

    @classmethod
    def as_dependant(cls) -> typing.Callable[..., 'QueryOptions']:

        def query_factory(**kwargs: typing.Any) -> cls:
            return cls(**kwargs)

        params = [
            inspect.Parameter(
                name='offset',
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
                default=fastapi.Query(
                    default=0,
                    alias='offset',
                    title="Offset",
                    description=(
                        "The offset that should be applied to the total result "
                        "set matching the search criteria."
                    )
                )
            ),
            inspect.Parameter(
                name='limit',
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
                default=fastapi.Query(
                    default=100,
                    alias='limit',
                    title="Limit",
                    description="Limit the number of objects in the result set."
                )
            ),
            inspect.Parameter(
                name='order_by',
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=typing.List[str],
                default=fastapi.Query(
                    default=[],
                    alias='order-by',
                    title="Order by",
                    description="Specifies attributes to order the results by."
                )
            ),
        ]
        params.extend(cls.get_factory_parameters())
        setattr(
            query_factory,
            '__signature__',
            inspect.Signature(parameters=params, return_annotation=cls)
        )

        return query_factory

    @classmethod
    def get_factory_parameters(cls) -> typing.List[inspect.Parameter]:
        """Update the signature of :meth:`query()` to reflect the
        attributes of the model.
        """
        params: typing.List[inspect.Parameter] = []
        for attname, field in cls.__fields__.items():
            if cls.is_reserved_field(attname):
                continue
            params.append(
                inspect.Parameter(
                    name=attname,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=field.field_info.extra.get('annotation'),
                    default=fastapi.Query(
                        default=field.default,
                        alias=field.alias,
                        title=field.field_info.title,
                        description=field.field_info.description
                    )
                )
            )
        return params


"""Declares :class:`ITokenIssuer`."""
from typing import Any

from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken

from .basegrant import BaseGrant
from .iclient import IClient
from .isubjectrepository import ISubjectRepository
from .servermetadata import ServerMetadata
from .tokenresponse import TokenResponse


class ITokenIssuer:
    __module__: str = 'cbra.ext.oauth2.types'

    #: Specifies the maximum age of an assertion.
    max_assertion_age: int = 30000

    #: Specifies the time-to-live of an access token if not otherwise
    #: specified, in seconds.
    default_ttl: int = 300

    #: The client that is handling the token request.
    client: IClient

    #: The concrete :class:`~cbra.ext.oauth2.ISubjectRepository`
    #: implementation that is used to lookup :term:`Subjects`.
    subjects: ISubjectRepository

    #: A :class:`~ckms.ext.oauth2.ServerMetadata` instance describing
    #: the authorization server that is issueing a token.
    metadata: ServerMetadata

    #: A :class:`ckms.jose.PayloadCodec` instance that knows how to
    #: decrypt JWE objects that were encrypted using the keys exposed
    #: at the JWKS URL.
    codec: PayloadCodec

    #: Indicates if a lookup of the JWKS of any issuer is allowed.
    allow_jwks_lookups: bool = True

    #: The set of trusted issuers for which the keys may be retrieved.
    trusted_issuers: set[str] = set()

    #: The URL of the local issuer i.e. the server. It is assumed that
    #: all tokens issued by this server have this value set to their
    #: issuer attribute.
    issuer: str

    def is_local_issued(self, claims: JSONWebToken) -> bool:
        raise NotImplementedError


"""Declares :class:`IssuerJWKS`."""
import urllib.parse

import httpx
import pydantic
from ckms.types import JSONWebKeySet

from cbra.exceptions import MisbehavingIssuer
from cbra.exceptions import UnreachableIssuer
from cbra.exceptions import UntrustedIssuer


class IssuerJWKS:
    """Maintains the keys for a remote issuer."""
    __module__: str = 'cbra.ext.jwks'
    keysets: dict[str, JSONWebKeySet]
    IssuerDoesNotExist: type[UntrustedIssuer] = UntrustedIssuer

    def __init__(self):
        self.keysets = {}

    async def get(
        self,
        issuer: str,
        trusted: set[str]
    ) -> JSONWebKeySet:
        if issuer not in self.keysets and issuer not in trusted:
            raise self.IssuerDoesNotExist(issuer)
        if issuer not in self.keysets:
            self.keysets[issuer] = await self.discover(issuer)
        return self.keysets[issuer]

    async def discover(self, issuer: str) -> JSONWebKeySet:
        """Discover the issuer's JSON Web Key Set (JWKS) using the
        OAuth 2.0/OpenID Connect server metadata.
        """
        p = urllib.parse.urlparse(issuer)
        base_url = f"{p.scheme or 'https'}://{p.netloc}"
        async with httpx.AsyncClient(base_url=base_url) as client:
            jwks_uri = None
            for path in [
                '/.well-known/openid-configuration',
                '/.well-known/oauth-authorization-server'
            ]:
                response = await self._get(
                    client=client,
                    issuer=issuer,
                    url=path
                )
                if response.status_code != 200:
                    continue
                metadata = response.json()
                jwks_uri = metadata.get('jwks_uri')
            if jwks_uri is None:
                raise MisbehavingIssuer(
                    detail=(
                        "The OAuth 2.0/OpenID Connect metadata could not be "
                        "retrieve from well-known URIs."
                    )
                )
            response = await self._get(
                client=client,
                issuer=issuer,
                url=jwks_uri
            )
            if response.status_code >= 400:
                raise MisbehavingIssuer(
                    detail="The `jwks_uri` returned a non-200 response."
                )
            try:
                jwks = JSONWebKeySet(**response.json())
            except (pydantic.ValidationError, TypeError, ValueError):
                raise MisbehavingIssuer(
                    detail=(
                        "The `jwks_uri` did not serve a JSON Web Key Set "
                        "(JWKS) or it could not be deserialized."
                    )
                )

        return jwks

    async def _get(
        self,
        client: httpx.AsyncClient,
        issuer: str,
        url: str
    ) -> httpx.Response:
        try:
            return await client.get(url) # type: ignore
        except httpx.RequestError:
            raise UnreachableIssuer


"""Declares :class:`PushedAuthorizationRequest`."""
from typing import Any

import pydantic
from ckms.types import JSONWebKey

from .baseauthorizationrequest import BaseAuthorizationRequest
from .iclient import IClient
from .iclientrepository import IClientRepository


_BaseParams = BaseAuthorizationRequest.body(
    exclude=['request', 'request_uri']
)


class JAR(_BaseParams):

    issuer: str = pydantic.Field(
        default=...,
        alias='iss',
        title="Issuer",
        description=(
            "The identifier of the client that is requesting "
            "authorization."
        )
    )

    audience: str | set[str] = pydantic.Field(
        default=...,
        alias='aud',
        title="Audience",
        description=(
            "Indicates the audience of the request. Must equal "
            "the `issuer` as specified by the authorization servers' "
            "metadata."
        )
    )

    access_token: str | None = pydantic.Field(
        default=None,
        title="Access token",
        description=(
            "An RFC 9068 access token that identifies the resource owner "
            "on during the **Authorization Code Flow**. This parameter "
            "**must not** be included in requests to the authorization "
            "endpoint; and may only be provided as part of a pushed "
            "authorization request."
        )
    )

    id_token: str | None = pydantic.Field(
        default=None,
        title="ID Token",
        description=(
            "An OpenID Connect ID Token that identifies the resource owner "
            "on during the **Authorization Code Flow**. This parameter "
            "**must not** be included in requests to the authorization "
            "endpoint; and may only be provided as part of a pushed "
            "authorization request."
        )
    )

    encryption_key: JSONWebKey | None = pydantic.Field(
        default=None,
        title="Encryption key",
        description=(
            "A public JSON Web Key (JWK) that is used to encrypt the response "
            "from the **Token Endpoint**."
        )
    )

    def get_claims(self) -> dict[str, Any]:
        return super().get_claims()

    async def get_client(self, clients: IClientRepository) -> IClient:
        """Return a :class:`~cbra.ext.oauth2.types.IClient` implementation
        determined from the request parameters.
        """
        return await clients.get(self.client_id)



# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
"""Declares :class:`IPrincipal`."""
from cbra.types import IPrincipal


__all__: list[str] = ['IPrincipal']


#pylint: skip-file
from .listqueryresult import ListQueryResult
from .namedquery import NamedQuery
from .pathfield import PathField
from .queryrunner import QueryRunner
from .requestquery import RequestQuery


__all__ = [
    'ListQueryResult',
    'NamedQuery',
    'PathField',
    'QueryRunner',
    'RequestQuery'
]



"""Declares :class:`Transaction`."""
import logging
import uuid
from typing import Any

import aorta
import fastapi

from .messagepublisher import MessagePublisher
from .params import ServerPublisher


class Transaction:
    """Encapsulates a transaction for :mod:`cbra` components, such as
    HTTP request handlers, webhooks, event listeners and commands handlers. The
    :class:`Transaction` object ensures that either all commands, events or
    tasks are published, or none of them.
    """
    logger: logging.Logger = logging.getLogger('uvicorn')
    _correlation_id: str | uuid.UUID | None = None
    _transaction_id: str
    _commands: list[aorta.Command]
    _events: list[aorta.Event]
    _publisher: aorta.MessagePublisher

    @property
    def correlation_id(self) -> str:
        if self._correlation_id is None:
            self._correlation_id = uuid.uuid4()
        return str(self._correlation_id)

    @correlation_id.setter
    def correlation_id(self, value: str | uuid.UUID):
        self._correlation_id = value

    def __init__(self, publisher: MessagePublisher = fastapi.Depends()):
        self._commands = []
        self._events = []
        self._publisher = publisher
        self._transaction_id = str(uuid.uuid4())

    def abort(self):
        if self._commands or self._events:
            self.logger.info(
                "Abort messaging transaction (correlationId: %s)",
                self.correlation_id
            )
            self._commands = []
            self._events = []

    def set_correlation_id(self, correlation_id: str | uuid.UUID | None) -> None:
        """Sets the correlation identifier for this transaction."""
        self._correlation_id = correlation_id or uuid.uuid4()

    async def commit(self):
        """Submits pending commands and events."""
        if self._commands or self._events:
            await self._publisher.publish(
                objects=self._commands + self._events, # type: ignore
                correlation_id=self.correlation_id
            )
            self.logger.info(
                "Commit messaging transaction (correlationId: %s)",
                self.correlation_id
            )

    def issue(self, command: aorta.Command):
        """Issue a command using the default command issuer."""
        if not self._commands:
            self.logger.info(
                "Begin messaging transaction (correlationId: %s)",
                self.correlation_id
            )
        self._commands.append(command)

    def publish(self, event: aorta.Event):
        """Publish an event using the default event publisher."""
        if not self._events:
            self.logger.info(
                "Begin messaging transaction (correlationId: %s)",
                self.correlation_id
            )
        self._events.append(event)

    async def __aenter__(self):
        return self

    async def __aexit__(
        self,
        cls: type[BaseException],
        exception: BaseException,
        traceback: Any
    ) -> bool:
        if exception is not None:
            self.abort()
            return False
        await self.commit()
        return False


class NullTransaction(Transaction):
    """A :class:`Transaction` implementation that does nothing."""

    def abort(self):
        pass

    async def commit(self):
        pass



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .iclient import IClient
from .isubject import ISubject


class IRefreshToken:
    __module__: str = 'cbra.ext.oauth2.types'

    async def generate(self, client: IClient, subject: ISubject) -> str | None:
        """Generate a new refresh token for the given `client`."""
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.routing
from fastapi.dependencies.utils import get_body_field
from fastapi.dependencies.utils import get_dependant

from .types import Request


RESPONSE_MODEL_PARAMS: set[str] = {
    'response_model',
    'response_model_include',
    'response_model_exclude',
    'response_model_by_alias',
    'response_model_exclude_unset',
    'response_model_exclude_defaults',
    'response_model_exclude_none',
}


class APIRoute(fastapi.routing.APIRoute):
    """A custom :class:`fastapi.routing.APIRoute` implementation
    that accomodates the rendering of request examples for multiple
    content types.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        # TODO: This here is a hack to remove unwanted stuff from the APIRoute
        # parameters.
        if kwargs.get('status_code') in (204, "204"):
            for param in RESPONSE_MODEL_PARAMS:
                kwargs.pop(param, None)
        super().__init__(*args, **kwargs)

        # Inspect the APIRoute.endpoint attribute to determine if we
        # need to do some magic the get a proper body model. Since
        # a number of handler functions does not declare the body
        # model using annotation, in order to be able to do explicit
        # parsing of the request body, it is not seen by FastAPI when
        # creating a route and rendering the OpenAPI schema. Thus, here
        # we look for some special types from which we can retrieve the
        # body model. This ensures that all models used by Endpoint
        # implementations are added to the schema.
        Model = getattr(self.endpoint, 'model', None)
        if Model is not None:
            # Create a fake dependant to trick get_body_field() into
            # getting the model.
            async def f(dto: Model) -> Any: pass # type: ignore
            self.body_field = get_body_field(
                dependant=get_dependant(path=self.path_format, call=f),
                name=self.unique_id
            )

    def get_route_handler(
        self
    ) -> Callable[[fastapi.Request], Coroutine[Any, Any, fastapi.Response]]:
        super_handler = super().get_route_handler()

        async def handler(request: fastapi.Request) -> fastapi.Response:
            return await super_handler(
                Request(request.scope, request.receive)
            )

        return handler


"""Declares :class:`DependantModel`."""
from __future__ import annotations
import typing
from inspect import isclass
from inspect import Parameter
from inspect import Signature

import fastapi
import fastapi.params
import pydantic


class DependantModel(pydantic.BaseModel):
    __module__: str = 'cbra.ext.oauth2.types'
    _exclude: typing.Set[str] = set()
    _source_class: typing.Type[pydantic.fields.FieldInfo] = fastapi.params.Query

    @classmethod
    def null(cls) -> None:
        return None

    @classmethod
    def as_dependant(cls, *args: typing.Any, **defaults: typing.Any) -> typing.Any:
        params: typing.List[Parameter] = []
        fields: typing.Dict[str, pydantic.fields.ModelField] = cls.__fields__
        for name, field in fields.items():
            if name in cls._exclude:
                continue
            kwargs = {
                k: getattr(field.field_info, k)
                for k in field.field_info.__slots__
            }
            kwargs.setdefault('alias', name)

            if name in defaults:
                default = defaults.pop(name)
            else:
                default = cls._source_class(**kwargs)
            if isclass(field.type_) and issubclass(field.type_, DependantModel):
                default = fastapi.Depends(field.type_.as_dependant())
            params.append(
                Parameter(
                    name=name,
                    kind=Parameter.POSITIONAL_OR_KEYWORD,
                    default=default,
                    annotation=field.type_
                )
            )

        async def f(**kwargs: typing.Any):
            return cls(**kwargs)
        
        f.__signature__ = Signature( # type: ignore
            parameters=params
        )
        return f

    @classmethod
    def fromquery(cls, **kwargs: typing.Dict[str, typing.Any]) -> typing.Any:
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeAlias

from .event import Event


__all__: list[str] = [
    'Event',
    'EventType',
    'OrderCreated',
    'OrderAllocated',
    'OrderClosed',
    'OrderStatusChanged',
    'OrderCompleted',
    'ProductChanged',
    'ProductFreeStockChanged',
    'ProductAssembledStockChanged',
    'ProductStockChanged',
]

OrderCreated: type[Event] = Event.new(
    name='OrderCreated',
    event='orders.created'
)

OrderAllocated: type[Event] = Event.new(
    name='OrderAllocated',
    event='orders.allocated'
)

OrderClosed: type[Event] = Event.new(
    name='OrderClosed',
    event='orders.closed'
)

OrderStatusChanged: type[Event] = Event.new(
    name='OrderStatusChanged',
    event='orders.status_changed'
)

OrderCompleted: type[Event] = Event.new(
    name='OrderCompleted',
    event='orders.completed'
)

ProductChanged: type[Event] = Event.new(
    name='ProductChanged',
    event='products.changed'
)

ProductFreeStockChanged: type[Event] = Event.new(
    name='ProductFreeStockChanged',
    event='products.free_stock_changed'
)

ProductAssembledStockChanged: type[Event] = Event.new(
    name='ProductAssembledStockChanged',
    event='products.assembled_stock_changed'
)

ProductStockChanged: type[Event] = Event.new(
    name='ProductStockChanged',
    event='products.stock_changed'
)

EventType: TypeAlias = (
    OrderCreated | OrderAllocated | OrderClosed |
    OrderStatusChanged | OrderCompleted |
    ProductChanged | ProductFreeStockChanged |
    ProductAssembledStockChanged |
    ProductStockChanged
)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets
from typing import Any

import pydantic
from ckms.core import parse_spec
from ckms.core import KeySpecification
from ckms.jose import PayloadCodec
from ckms.utils import b64decode

from .iclient import IClient
from .irefreshtoken import IRefreshToken
from .isubject import ISubject


KEY_LENGTH: int = 64


class RefreshToken(pydantic.BaseModel, IRefreshToken):
    id: int | None = None
    sub: int | str
    client_id: str
    secret: str
    created: datetime.datetime
    exchanged: datetime.datetime
    scope: list[str]

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        values.setdefault('secret', secrets.token_urlsafe(KEY_LENGTH))
        return values

    @classmethod
    def new(
        cls,
        client_id: str,
        sub: int | str,
        scope: list[str] | set[str]
    ) -> 'RefreshToken':
        now = datetime.datetime.utcnow()
        return cls.parse_obj({
            'client_id': client_id,
            'sub': sub,
            'secret': secrets.token_urlsafe(KEY_LENGTH),
            'created': now,
            'exchanged': now,
            'scope': scope
        })

    @property
    def key(self) -> KeySpecification:
        return parse_spec({
            'provider': 'local',
            'kty': 'oct',
            'alg': 'HS256',
            'use': 'sig',
            'key': {'cek': b64decode(self.secret)}
        })

    async def get_codec(self) -> PayloadCodec:
        return PayloadCodec(signing_keys=[await self.key])

    async def generate(self, client: IClient, subject: ISubject) -> str | None:
        assert self.id is not None # nosec
        self.exchanged = datetime.datetime.utcnow()
        self.secret = secrets.token_urlsafe(KEY_LENGTH)
        return await client.create_refresh_token(
            codec=await self.get_codec(),
            subject=subject,
            token_id=self.id
        )


# pylint: skip-file
# The code in this file is taken from Django and Django RESt Framework.
"""
Handling of media types, as found in HTTP Content-Type and Accept headers.

See https://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7
"""
import typing

HTTP_HEADER_ENCODING = 'iso-8859-1'


def media_type_matches(lhs, rhs):
    """
    Returns ``True`` if the media type in the first argument <= the
    media type in the second argument.  The media types are strings
    as described by the HTTP spec.

    Valid media type strings include:

    'application/json; indent=4'
    'application/json'
    'text/*'
    '*/*'
    """
    lhs = _MediaType(lhs)
    rhs = _MediaType(rhs)
    return lhs.match(rhs)


def order_by_precedence(media_type_lst) -> typing.List[str]: # pragma: no cover
    """
    Returns a list of sets of media type strings, ordered by precedence.
    Precedence is determined by how specific a media type is:

    3. 'type/subtype; param=val'
    2. 'type/subtype'
    1. 'type/*'
    0. '*/*'
    """
    ret = [set(), set(), set(), set()]
    for media_type in media_type_lst:
        precedence = _MediaType(media_type).precedence
        ret[3 - precedence].add(media_type)
    return [media_types for media_types in ret if media_types]


class _MediaType: # pragma: no cover
    def __init__(self, media_type_str):
        self.orig = '' if (media_type_str is None) else media_type_str
        self.full_type, self.params = parse_header(self.orig.encode(HTTP_HEADER_ENCODING))
        self.main_type, sep, self.sub_type = self.full_type.partition('/')

    def match(self, other):
        """Return true if this MediaType satisfies the given MediaType."""
        for key in self.params:
            if key != 'q' and other.params.get(key, None) != self.params.get(key, None):
                return False

        if self.sub_type != '*' and other.sub_type != '*' and other.sub_type != self.sub_type:
            return False

        if self.main_type != '*' and other.main_type != '*' and other.main_type != self.main_type:
            return False

        return True

    @property
    def precedence(self):
        """
        Return a precedence level from 0-3 for the media type given how specific it is.
        """
        if self.main_type == '*':
            return 0
        elif self.sub_type == '*':
            return 1
        elif not self.params or list(self.params) == ['q']:
            return 2
        return 3

    def __str__(self):
        ret = "%s/%s" % (self.main_type, self.sub_type)
        for key, val in self.params.items():
            ret += "; %s=%s" % (key, val.decode('ascii'))
        return ret


def parse_header(line: bytes) -> typing.Dict[str, bytes]: # pragma: no cover
    """
    Parse the header into a key-value.
    Input (line): bytes, output: str for key/name, bytes for values which
    will be decoded later.
    """
    plist = _parse_header_params(b';' + line)
    key = plist.pop(0).lower().decode('ascii')
    pdict = {}
    for p in plist:
        i = p.find(b'=')
        if i >= 0:
            has_encoding = False
            name = p[:i].strip().lower().decode('ascii')
            if name.endswith('*'):
                # Lang/encoding embedded in the value (like "filename*=UTF-8''file.ext")
                # https://tools.ietf.org/html/rfc2231#section-4
                name = name[:-1]
                if p.count(b"'") == 2:
                    has_encoding = True
            value = p[i + 1:].strip()
            if len(value) >= 2 and value[:1] == value[-1:] == b'"':
                value = value[1:-1]
                value = value.replace(b'\\\\', b'\\').replace(b'\\"', b'"')
            if has_encoding:
                encoding, lang, value = value.split(b"'")
                value = unquote(value.decode(), encoding=encoding.decode())
            pdict[name] = value
    return key, pdict


def _parse_header_params(s): # pragma: no cover
    plist = []
    while s[:1] == b';':
        s = s[1:]
        end = s.find(b';')
        while end > 0 and s.count(b'"', 0, end) % 2:
            end = s.find(b';', end + 1)
        if end < 0:
            end = len(s)
        f = s[:end]
        plist.append(f.strip())
        s = s[end:]
    return plist



# type: ignore
"""Declares :class:`ClientConfig`."""
import datetime
import logging
import secrets
import typing
from typing import Any, Literal

import httpx
import pydantic
from cbra.exceptions import UpstreamServiceNotAvailable
from ckms.core.models import JSONWebSignature
from ckms.types import IKeychain
from ckms.types import JSONWebKey
from ckms.types import JSONWebKeySet
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload
from ckms.jose import PayloadCodec
from unimatrix.exceptions import CanonicalException

from cbra.utils import current_timestamp
from .baseclient import BaseClient
from .clientsubject import ClientSubject
from .exceptions import Error
from .exceptions import MissingRedirectURL
from .exceptions import RedirectForbidden
from .types import ISubject
from .types import GrantType
from .types import RedirectURL
from .types import RefreshTokenPolicy
from .types import ResponseType


logger: logging.Logger = logging.getLogger('uvicorn')
JWKSLookupFailed: UpstreamServiceNotAvailable = UpstreamServiceNotAvailable(5)


class ClientConfig(BaseClient):
    InvalidCredential: type[Exception] = type('InvalidCredential', (Exception,), {})

    allow_key_discovery: bool = False

    allow_multiple_audiences: bool = False

    allowed_audience: typing.Set[str] = set()

    allowed_origins: typing.Set[str] = set()

    client_id: str

    client_secret: str | None = None

    default_redirect_url: RedirectURL | None = None

    first_party: bool = False

    grant_types_supported: list[GrantType] = []

    id_token_ttl: int = 3600

    jwks_uri: str | None = None

    keys: list[JSONWebKey] = []

    redirect_uris: list[RedirectURL] = []

    refresh_token_policy: RefreshTokenPolicy = RefreshTokenPolicy.rolling

    refresh_token_ttl: int = 0

    response_modes: set[str] = set()

    response_types: set[ResponseType] = set()

    scope: set[str] = set()

    subject_type: Literal["public", "pairwise"] = "pairwise"

    @property
    def jwks(self) -> JSONWebKeySet:
        return JSONWebKeySet(keys=self.keys)

    def __init__(self, **kwargs: typing.Any):
        redirect_uris = kwargs.setdefault('redirect_uris', [])
        if 'default_redirect_url' in kwargs:
            redirect_uris.append(kwargs['default_redirect_url'])
        super().__init__(**kwargs)

    async def get_codec(self, codec: PayloadCodec) -> PayloadCodec:
        """Return a new :class:`ckms.jose.PayloadCodec` configured to use
        the client keys for encryption.
        """
        jwks = await self.get_jwks()
        return PayloadCodec(
            signer=codec.signer,
            verifier=JSONWebKeySet(keys=[
                jwk for jwk in jwks.keys 
                if jwk.use == 'sig'
            ]),
            encryption_keys=[
                jwk for jwk in jwks.keys
                if jwk.use == 'enc'
            ]
        )

    async def get_jwks(self) -> JSONWebKeySet:
        """Return a :class:`ckms.types.JSONWebKeySet` instance holding the
        signing and encryption keys of the client.
        """
        jwks = self.jwks
        if self.jwks_uri:
            jwks = await self.import_jwks()
        return jwks

    def allows_audience(self, issuer: str, audience: typing.Set[str]) -> bool:
        """Return a boolean indicating if the client allows the issueing
        of access token for the given audiences set `audience`.
        """
        return (audience <= (self.allowed_audience | {issuer}))

    def allows_grant_type(self, grant_type: GrantType) -> bool:
        """Return a boolean indicating if the client allows the
        given grant type.
        """
        return grant_type.value in self.grant_types_supported

    def allows_origin(self, origin: str | None) -> bool:
        """Return a boolean indicating if the client allows the given origin in the
        context of Cross-Origin Resource Sharing (CORS).
        """
        return origin is None or origin in self.allowed_origins

    def allows_response_mode(self, response_mode: Any) -> bool:
        """Return a boolean indicating if the client allows the given response
        mode.
        """
        return response_mode in self.response_modes

    def allows_response_type(self, response_type: ResponseType) -> bool:
        """Return a boolean indicating if the client allows the given response
        type.
        """
        return response_type in self.response_types

    def allows_scope(self, scope: set[str]) -> bool:
        """Return a boolean indicating if the client allows the given scope."""
        return bool(set(scope) <= self.scope)

    def as_subject(self) -> ISubject:
        """Return a :class:`~cbra.ext.oauth2.types.ISubject` implementation
        representing the client itself. For use with the client credentials
        grant.
        """
        return ClientSubject(client_id=self.client_id)

    def can_issue_multiple(self) -> bool:
        """Return a boolean indicating if the client allows issueing
        access tokens to multiple audiences.
        """
        return self.allow_multiple_audiences

    def can_redirect(self, url: typing.Union[str, RedirectURL]) -> bool:
        """Return a boolean indicating if the client allows
        redirects to the given `url`.
        """
        return url in self.redirect_uris

    def get_default_redirect(self) -> str:
        """Return the default redirect URL, if specified."""
        return self.default_redirect_url

    def get_encryption_keys(self) -> list[JSONWebKey]:
        """Return the encryption keys used by the client."""
        return [x for x in self.keys if x.use == "enc"]

    def get_id_token_claims(self, now: int) -> dict[str, Any]:
        return {
            'aud': [self.client_id],
            'azp': self.client_id,
            'exp': now + self.id_token_ttl
        }

    def get_redirect_url(
        self,
        url: str | RedirectURL | None = None,
        fatal: bool = True
    ) -> RedirectURL | None:
        """Return a :class:`RedirectURL` instance that the server is
        allowed to redirect to.
        """
        # If there is no URL or a single redirect URL, then the
        # client is allowed to omit the redirect_uri parameter.
        if not url and len(self.redirect_uris) == 1:
            return self.redirect_uris[0]

        # If multiple redirect URIs have been registered, the client
        # MUST include a redirect URI with the authorization request
        # using the "redirect_uri" request parameter (OAuth 2.1 draft,
        # 2.3.3).
        try:
            if url is None and len(self.redirect_uris) > 1:
                raise MissingRedirectURL
            if url is None and self.redirect_uris:
                url = self.redirect_uris[0]
            if not self.redirect_uris:
                raise RedirectForbidden

            if not isinstance(url, RedirectURL):
                url = RedirectURL.validate(url)
            if not self.can_redirect(url):
                raise RedirectForbidden
        except Error:
            if fatal:
                raise
            url = None
        return url

    def get_refresh_token_expires(self, now: int) -> int:
        """Return a :class:`datetime.datetime` instance indicating the
        Time-To-Live (TTL) of a refresh token.
        """
        return now + self.refresh_token_ttl

    def get_subject_id(self, subject: ISubject) -> int:
        """Return the subject identifier to add to an access token or ID
        token.
        """
        return subject.get_identifier(public=self.subject_type == "public")

    async def create_refresh_token(
        self,
        codec: PayloadCodec,
        subject: ISubject,
        token_id: int
    ) -> str:
        now = current_timestamp()
        exp = self.get_refresh_token_expires(now)
        return await codec.encode(
            payload={
                'aud': self.client_id,
                'exp': exp,
                'iat': now,
                'iss': token_id,
                'jti': secrets.token_urlsafe(24),
                'nbf': now
            },
            content_type="rt+jwt"
        )

    async def import_jwks(self) -> JSONWebKeySet:
        """Retrieve the JSON Web Key Set (JWKS) that is specified by the
        :attr:`jwks_uri` attribute.

        Raises:
            :class:`~cbra.exceptions.UpstreamServiceNotAvaialble`: unable
                to lookup the JWKS through :attr:`jwks_uri`.
        """
        jwks = self.jwks
        if not jwks.keys and self.jwks_uri is not None:
            async with httpx.AsyncClient(timeout=5.0) as client:
                logger.info(
                    "Obtaining client (id: %s) keys from %s",
                    self.client_id, self.jwks_uri
                )
                try:
                    response = await client.get(self.jwks_uri)
                    response.raise_for_status()
                except (httpx.RequestError, httpx.HTTPError):
                    # Occurs when there is a network problem i.e. connection timeout,
                    # DNS not resolved
                    logger.critical(
                        "Unable to lookup client (id: %s) JWKS from %s",
                        self.client_id, self.jwks_uri
                    )
                    raise JWKSLookupFailed
                try:
                    jwks = JSONWebKeySet.parse_obj(response.json())
                    keys = jwks.keys
                    if not jwks.keys or not [x for x in jwks.keys if x.use == 'sig']:
                        logger.critical(
                            "No signing keys in client (id: %s) JWKS retrieved from %s",
                            self.client_id, self.jwks_uri
                        )
                        raise UpstreamServiceNotAvailable(
                            code="INVALID_JWKS",
                            message="The JSON Web Key Set (JWKS) did not contain any signing keys."
                        )
                    self.keys = jwks.keys
                except (pydantic.ValidationError, TypeError, ValueError) as e:
                    logger.critical("Unable to parse JWKS from %s", self.jwks_uri)
                    raise UpstreamServiceNotAvailable(
                        code="INVALID_JWKS_FORMAT",
                        message=(
                            f"The JSON Web Key Set (JWKS) retrieved from {self.jwks_uri} "
                            "could not be interpreted as JSON or the schema of the JSON "
                            "object was invalid."
                        )
                    )
        return jwks

    def is_first_party(self) -> bool:
        """Return a boolean indicating if this is a first-party client
        that does not have to obtain consent.
        """
        return self.first_party

    async def issue_token(
        self,
        codec: PayloadCodec,
        using: str,
        issuer: str,
        audience: typing.Union[str, typing.Set[str]],
        subject: ISubject,
        ttl: int,
        scope: typing.Set[str]
    ) -> str:
        """Issue an access token encoded as a JSON Web Token (JWT)
        as defined in :rfc:`9068`. Return a string containing the
        token.
        """
        claims = {
            'aud': list(audience),
            'client_id': self.client_id,
            'iss': issuer,
            'jti': secrets.token_urlsafe(24),
            'scope': str.join(' ', sorted(scope)),
            'sub': self.get_subject_id(subject),
        }
        if scope:
            claims['scope'] = ' '.join(sorted(x for x in scope))

        # The scope claim can be removed if it is empty because "if an
        # authorization request includes a scope parameter, the
        # corresponding issued JWT access token SHOULD include a "scope"
        # claim as defined in Section 4.2 of [RFC8693]." (RFC 9068, 2.2.3)
        if not scope:
            claims.pop('scope')

        now = claims['iat'] = current_timestamp()
        claims.update({
            'exp': now + ttl,
            'nbf': now,
        })
        return await codec.encode(
            JSONWebToken(**claims),
            content_type="at+jwt",
            signers=[using]
        )

    def is_confidential(self) -> bool:
        """Return a boolean indicating if the client is confidential."""
        return bool(self.jwks_uri or self.keys or self.client_secret)

    def is_public(self) -> bool:
        """Return a boolean indicating if the client is public."""
        return self.client_secret is None and not self.keys

    async def verify_jws(
        self,
        jws: JSONWebSignature
    ) -> bool:
        """Verify a :class:`~ckms.jose.JSONWebSignature` using the public
        keys associated to the client.
        """
        return await jws.verify(await self.get_jwks())

    async def verify_jwt(
        self,
        token: typing.Union[bytes, str],
        decrypter: typing.Optional[IKeychain],
        compact_only: bool = False,
        audience: str | None = None
    ) -> JSONWebToken:
        """Verifies a JSON Web Token (JWT) using the public keys of the
        client.

        Args:
            token (byte-sequence or string): the JWT to deserialize and
                verify.
            decrypter: a :class:`~IKeychain` implementation that has the
                private keys to decrypt, if the payload is a JWE.
            compact_only (bool): indicates if only compact deserialization
                is allowed. If `compact_only` is ``True`` and the token
                is not compact serialized, an exception is raised.

        Returns
            A :class:`dict` containing the claims in the JWT.
        """
        codec = PayloadCodec(decrypter=decrypter, verifier=await self.get_jwks())
        if isinstance(token, str): # pragma: no cover
            token = str.encode(token, 'ascii')
        jwt = await codec.decode(token, accept="jwt")
        if not isinstance(jwt, JSONWebToken):
            raise MalformedPayload(message="A JSON Web Token (JWT) is expected.")
        jwt.verify(audience={audience} if audience else set())
        return jwt


"""Declares common fields used in OAuth 2.0."""
import typing

import pydantic


def ResourceField():
    return pydantic.Field(
        alias='resource',
        default=set(),
        title="Resource",
        description=(
            "Indicates the target service or resource to which access is "
            "being requested.  Its value MUST be an absolute URI, as "
            "specified by Section 4.3 of RFC 3986.  The URI MUST NOT "
            "include a fragment component. It SHOULD NOT include a query "
            "component, but it is recognized that there are cases that "
            "make a query component a useful and necessary part of the "
            "resource parameter, such as when one or more query parameters "
            "are used to scope requests to an application.  The `resource` "
            "parameter URI value is an identifier representing the identity "
            "of the resource, which MAY be a locator that corresponds to a "
            "network-addressable location where the target resource is "
            "hosted.  Multiple `resource` parameters MAY be used to indicate "
            "that the requested token is intended to be used at multiple "
            "resources."
        ),
        example="https://hello.unimatrixapis.com"
    )


def GrantTypeField(
    cls: typing.Callable[..., typing.Any] = pydantic.Field,
    default: typing.Any =...
) -> typing.Any:
    return cls(
        default=default,
        title="Grant type",
        description=(
            "Specifies the grant type that is requested by the client."
        ),
        example="authorization_code"
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .basegrant import BaseGrant
from .spaceseparatedlist import SpaceSeparatedList


class ScopedGrant(BaseGrant):
    scope: SpaceSeparatedList | set[str] | None = pydantic.Field(
        default=[],
        title="Scope",
        description=(
            "The requested scope that was previously granted to the subject "
            "or client."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
import base64
import json
from typing import Any

import pytest
from httpx import AsyncClient


class BaseMessageTest:
    event_type: str
    resource_type: str

    def get_request_params(self) -> dict[str, Any]:
        raise NotImplementedError

    def test_content_is_json_parseable(self) -> None:
        params = self.get_request_params()
        json.loads(bytes.decode(params['content']))

    @pytest.mark.asyncio
    async def test_accept_event(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 200, response.text
        assert dto.get('status') in {'accepted', 'dropped'}, dto

    @pytest.mark.asyncio
    async def test_disallowed_event_is_rejected(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        params['headers']['x-wc-webhook-event'] = 'foo'
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 200, response.text
        assert dto.get('status') == 'dropped'

    @pytest.mark.asyncio
    async def test_missing_signature(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        params['headers'].pop('x-wc-webhook-signature')
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 403, response.text
        assert dto['spec']['code'] == 'WEBHOOK_EVENT_REJECTED'
        assert dto['spec']['message'] == "The request signature is not valid."


    @pytest.mark.asyncio
    async def test_malformed_signature(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        params['headers']['x-wc-webhook-signature'] = 'foo'
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 403, response.text
        assert dto['spec']['code'] == 'WEBHOOK_EVENT_REJECTED'
        assert dto['spec']['message'] == "The signature is malformed."


    @pytest.mark.asyncio
    async def test_mismatching_signature(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        params['headers']['x-wc-webhook-signature'] = base64.b64encode(b'foo')
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 403, response.text
        assert dto['spec']['code'] == 'WEBHOOK_EVENT_REJECTED'
        assert dto['spec']['message'] == "The request signature is not valid."


# pylint: disable=E0202,W0212
"""Declares :class:`Dependency`."""
import asyncio
import inspect
import typing

import fastapi.params


class Dependency(fastapi.params.Depends):
    """Like :class:`fastapi.Depends`, but exposes a method to get
    the dependency. This allows the subclassing of :class:`Dependency` for
    specific use cases and dependency injection strategies.

    .. literalinclude:: ../../../examples/dependency-injection-dependency.py
      :language: python

    The `dependency` argument is used to inject the function that resolves
    the dependency. If it is omitted, then :meth:`get_resolver()` is invoked
    when the :class:`Dependency` is being called.

    Subclasses may override the :meth:`get_resolver()` method to implement
    additional logic when resolving the callable that is injected.

    .. automethod:: resolve()

    .. automethod:: get_resolver()
    """
    __module__: str = 'cbra.ext.ioc'
    _is_coroutine = asyncio.coroutines._is_coroutine

    @property
    def __signature__(self) -> inspect.Signature:
        return self.get_signature()

    def __init__(self,
        dependency: typing.Optional[typing.Callable[..., typing.Any]] = None,
        use_cache: bool = True
    ):
        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec
        if dependency is not None:
            self.resolve = dependency
        super().__init__(
            dependency=self,
            use_cache=use_cache
        )

    def get_resolver(self) -> typing.Callable:
        """Return the callable that is used to resolve the dependency."""
        return self.resolve

    def get_signature(self) -> inspect.Signature:
        """Return a :class:`inspect.Signature` instance representing the
        call sigature of the dependency resolver.
        """
        return inspect.signature(self.get_resolver())

    async def resolve(self):
        """Resolve the dependency. The default implementation always raises
        :exc:`NotImplementedError`.
        """
        raise NotImplementedError

    async def __call__(self, *args, **kwargs) -> typing.Any:
        resolver = self.get_resolver()
        if inspect.isasyncgenfunction(resolver)\
        or inspect.isgeneratorfunction(resolver):
            raise ValueError(
                "Dependencies must not be (async) generators."
            )
        return await resolver(*args, **kwargs)



"""Declares :class:`NullParser`."""
import typing

import fastapi

from cbra.types import IParser


class NullParser(IParser):
    __module__: str = 'cbra.parser'

    async def parse(self,
        request: fastapi.Request,
        media_type: typing.Optional[str] = None,
        parser_context: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> typing.Any:
        return None


# pylint: skip-file
from .default import DefaultContentNegotiation
from .json import JSONDefaultContentNegotiation
from .null import NullContentNegotiation
from .nullresponse import NullResponseContentNegotiation


__all__ = [
    'DefaultContentNegotiation',
    'JSONDefaultContentNegotiation',
    'NullContentNegotiation',
    'NullResponseContentNegotiation',
]



"""Declares :class:`TokenExchangeResponse`"""
import typing

import pydantic

from .tokentype import TokenType


class TokenExchangeResponse(pydantic.BaseModel):
    __module__: str = 'cbra.ext.oauth2'

    access_token: str = pydantic.Field(
        default=...,
        description=(
            "The security token issued by the authorization server "
            "in response to the token exchange request."
        )
    )

    issued_token_type: TokenType = pydantic.Field(
        default=...,
        title="Issued token type",
        description=(
                "An identifier for the representation of the issued security token."
        ),
        example="urn:ietf:params:oauth:token-type:jwt"
    )

    token_type: str = pydantic.Field(
        default=...,
        description=(
            "A case-insensitive value specifying the method of using "
            "the access token issued. It provides the client with "
            "information about how to utilize the access token to "
            "access protected resources.  For example, a value of "
            "`Bearer` indicates that the issued security token is "
            "a bearer token and the client can simply present it "
            "as is without any additional proof of eligibility "
            "beyond the contents of the token itself.  Note that "
            "the meaning of this parameter is different from the "
            "meaning of the `issued_token_type` parameter, which "
            "declares the representation of the issued security "
            "token; the term `token type` is more typically used "
            "to mean the structural or syntactical representation "
            "of the security token, as it is in all `*_token_type` "
            "parameters.  If the issued token is not an access "
            "token or usable as an access token, then the `token_type` "
            "value `N_A` is used to indicate that an OAuth 2.0 "
            "`token_type` identifier is not applicable in that context."
        ),
        example='Bearer'
    )

    expires_in: typing.Optional[int] = pydantic.Field(
        default=None,
        title="Expires in",
        description=(
            "The validity lifetime, in seconds, of the token issued by the "
            "authorization server.  Oftentimes, the client will not have "
            "the inclination or capability to inspect the content of the "
            "token, and this parameter provides a consistent and token-type "
            "agnostic indication of how long the token can be expected to "
            "be valid.  For example, the value 1800 denotes that the "
            "token will expire in thirty minutes from the time the "
            "response was generated."
        ),
        example=1800
    )

    scope: typing.Optional[str] = pydantic.Field(
        default=None,
        description=(
            "Present if the scope of the issued security token is different "
            "to the scope requested by the client; otherwise it is absent."
        ),
        example="foo bar baz"
    )

    refresh_token: typing.Optional[str] = pydantic.Field(
        default=None,
        description=(
            "A refresh token will typically not be issued when the exchange "
            "is of one temporary credential (the subject_token) for a "
            "different temporary credential (the issued token) for use "
            "in some other context.  A refresh token can be issued in "
            "cases where the client of the token exchange needs the "
            "ability to access a resource even when the original "
            "credential is no longer valid (e.g., user-not-present or "
            "offline scenarios where there is no longer any user "
            "entertaining an active session with the client). Profiles "
            "or deployments of this specification should clearly document "
            "the conditions under which a client should expect a refresh "
            "token in response to `urn:ietf:params:oauth:grant-type:token-exchange`"
            " grant type requests."
        )
    )

    class Config:
        title = "Token Exchange Response"
        exclude_defaults = True


"""Declares :class:`Action`."""
from email.policy import default
import inspect
import types
import typing
from typing import Any
from typing import Callable
from typing import TypeAlias

import pydantic
from cbra.headers import HEADERS_RESPONSE_BODY
from cbra.headers import CORS_HEADERS
from cbra.headers import DIGEST_SCHEMA
from cbra.types import IEndpoint
from .iresource import IResource
from .pathparameter import PathParameter
from .resourceoptions import ResourceOptions


ModelType: TypeAlias = type[pydantic.BaseModel] | types.UnionType


class Action:
    __module__: str = 'cbra.resource'
    summary = {
        'create'    : "Create {article} {name}",
        'retrieve'  : "Retrieve {article} {name}",
        'update'    : "Update (patch) {article} {name}",
        'replace'   : "Replace {article} {name}",
        'delete'    : "Delete {article} {name}",
        'list'      : "List {pluralname}",
        'purge'     : "Delete all {pluralname}",
    }
    default_actions: typing.Dict[str, typing.Tuple[str, bool]] = {
        'create'    : ('POST', False),
        'list'      : ('GET', False),
        'purge'     : ('DELETE', False),
        'retrieve'  : ('GET', True),
        'update'    : ('PATCH', True),
        'replace'   : ('PUT', True),
        'delete'    : ('DELETE', True),
    }
    is_detail: bool
    name: str
    method: str
    path_parameter: PathParameter
    response_descriptions: typing.Dict[str, str] = {
        'create'    : "The created **{name}** object or a reference to its location.",
        'retrieve'  : "The **{name}** object specified by the path parameter(s).",
        'purge'     : "The list of deleted **{name}** objects.",
        'delete'    : "The deleted **{name}** object specified by the path parameter(s).",
        'list'      : "The list of **{pluralname}** objects matching the search criterion.",
        'update'    : "The updated **{name}** object specified by the path parameter(s).",
        'replace'   : "The replaced **{name}** object specified by the path parameter(s).",
    }
    subpath: typing.Optional[str]
    queryable_methods = {"GET", "DELETE"}

    @staticmethod
    def creates_resource(action: str) -> bool:
        return action == "create"

    @staticmethod
    def model_from_signature(func: Callable[..., Any]) -> ModelType | None:
        sig = inspect.signature(func)
        args = list(sig.parameters.values())
        if len(args) == 1: # No request body model
            return None
        model: None | ModelType = None
        if inspect.isclass(args[1].annotation)\
        and issubclass(args[1].annotation, (pydantic.BaseModel, types.UnionType)):
            model = args[1].annotation # type: ignore
        return model

    @classmethod
    def fromclass(
        cls,
        resource_class: typing.Type[IResource]
    ) -> typing.Generator['Action', None, None]:

        path_parameter = typing.cast(PathParameter, resource_class.path_parameter)
        detail_methods: typing.Set[str] = set()
        list_methods: typing.Set[str] = set()
        for attname, member in inspect.getmembers(resource_class):
            if attname not in cls.default_actions\
            and (not hasattr(member, 'action') or not inspect.isfunction(member)):
                continue
            if hasattr(member, 'action'):
                action_name = member.action.pop('name')
                handler: typing.Type[IEndpoint] = resource_class.new(
                    action=action_name,
                    cors_policy=resource_class.cors_policy,
                    handle=member,
                    method="POST",
                    model=cls.model_from_signature(member),
                    name=resource_class.name,
                    **member.action
                )
                action = Action(
                    name=action_name,
                    handler=typing.cast(typing.Type[IResource], handler),
                    method=handler.method,
                    is_detail=True,
                    path_parameter=path_parameter
                )
                yield action
                yield action.as_options(
                    summary=member.action.get('summary') or action_name,
                    methods=["POST"],
                    name=action_name,
                    subpath=action_name
                )
                continue
            if attname in cls.default_actions:
                method, is_detail = cls.default_actions[attname]
                response_description = cls.response_descriptions.get(attname, "Success")\
                    .format(
                        article=resource_class.name_article,
                        name=resource_class.verbose_name,
                        pluralname=resource_class.verbose_name_plural
                    )


                # Hack: determine some parameter if there is no response.
                response_model = resource_class.model
                default_response_code = 201 if cls.creates_resource(attname) else 200
                if inspect.signature(member).return_annotation is None:
                    default_response_code = 204
                handler: typing.Type[IEndpoint] = resource_class.new(
                    action=attname,
                    cors_policy=resource_class.cors_policy,
                    default_response_code=default_response_code,
                    handle=member,
                    is_detail=is_detail,
                    method=method,
                    model=response_model,
                    name=resource_class.name,
                    queryable=not is_detail and method in cls.queryable_methods,
                    response_description=getattr(member, 'response_description', response_description),
                    searchable=attname in resource_class.filter_actions,
                    summary=getattr(member, "summary", None)
                )
                yield Action(
                    name=attname,
                    handler=typing.cast(typing.Type[IResource], handler),
                    method=method,
                    is_detail=is_detail,
                    path_parameter=path_parameter
                )
                detail_methods.add(method)\
                    if is_detail else\
                    list_methods.add(method)

        if list_methods:
            yield Action(
                name='options',
                path_parameter=path_parameter,
                method='OPTIONS',
                handler=typing.cast(
                    typing.Type[IResource],
                    ResourceOptions.new(
                        allowed_methods=list_methods,
                        cors_policy=resource_class.cors_policy,
                        document=resource_class.document,
                        is_detail=False,
                        model=resource_class.model,
                        path_parameters=resource_class.get_path_signature(False),
                        name=resource_class.name,
                        pluralname=resource_class.pluralname,
                        summary="Collection endpoint options and CORS policy",
                        verbose_name=resource_class.verbose_name,
                        verbose_name_plural=resource_class.verbose_name_plural
                    )
                ),
                is_detail=False
            )

        if detail_methods:
            assert path_parameter.signature_parameter is not None # nosec
            yield Action(
                name='options',
                path_parameter=path_parameter,
                method='OPTIONS',
                handler=typing.cast(
                    typing.Type[IResource],
                    ResourceOptions.new(
                        allowed_methods=detail_methods,
                        cors_policy=resource_class.cors_policy,
                        document=resource_class.document,
                        is_detail=True,
                        model=resource_class.model,
                        path_parameters=resource_class.get_path_signature(True),
                        name=resource_class.name,
                        pluralname=resource_class.pluralname,
                        summary="Detail endpoint options and CORS policy",
                        verbose_name=resource_class.verbose_name,
                        verbose_name_plural=resource_class.verbose_name_plural
                    )
                ),
                is_detail=True
            )

    def __init__(
        self,
        *,
        name: str,
        handler: typing.Type[IResource],
        method: str,
        is_detail: bool,
        path_parameter: PathParameter,
        subpath: typing.Optional[str] = None,
    ):
        self.name = name
        self.handler = handler
        self.method = method
        self.is_detail = is_detail
        self.subpath = subpath
        self.path_parameter = path_parameter

    def as_options(
        self,
        *,
        summary: str,
        methods: list[str],
        name: str = 'options',
        subpath: str | None = None,
    ) -> 'Action':
        """Return the options action for this action."""
        cls = self.handler
        return Action(
            name=name,
            path_parameter=self.path_parameter,
            method='OPTIONS',
            handler=typing.cast(
                typing.Type[IResource],
                ResourceOptions.new(
                    allowed_methods=methods,
                    cors_policy=cls.cors_policy,
                    document=cls.document,
                    is_detail=self.is_detail,
                    path_parameters=cls.get_path_signature(self.is_detail),
                    name=cls.name,
                    pluralname=cls.pluralname,
                    summary=summary,
                    verbose_name=cls.verbose_name,
                    verbose_name_plural=cls.verbose_name_plural
                )
            ),
            is_detail=self.is_detail,
            subpath=subpath
        )

    def add_to_router(
        self,
        app: IEndpoint.RouterType,
        base_path: str
    ) -> None:
        path = base_path if not self.is_detail else f'{base_path}/{{id}}'
        subpath = None
        if (self.name not in self.default_actions)\
        and (self.name != 'options'):
            # TODO: Find a cleaner way to construct the path.
            subpath = self.subpath or self.name

        path = self.path_parameter.get_path(
            base_path=base_path,
            subpath=subpath,
            is_detail=self.is_detail
        )
        self.handler.add_to_router(
            app=app,
            base_path=path,
            method=self.method,
            request_handler=...,
            **self.get_app_parameters(self.handler)
        )

    def get_app_parameters(
        self,
        resource_class: typing.Type[IResource]
    ) -> typing.Dict[str, typing.Any]:
        params: typing.Dict[str, typing.Any] = {
            'summary': resource_class.summary,
            'tags': [resource_class.verbose_name_plural],
            'openapi_extra': self.get_openapi_schema(resource_class)
        }
        if self.name != 'options':
            params['name'] = str.lower(f'{self.handler.name}.{self.name}')

        description = getattr(resource_class.handle, '__doc__', None)
        if description:
            params['description'] = description
        if self.name in self.summary and not resource_class.summary:
            params['summary'] = self.summary[self.name].format(
                article=resource_class.name_article,
                name=resource_class.verbose_name or str.title(resource_class.name),
                pluralname=resource_class.verbose_name_plural\
                    or str.title(resource_class.pluralname)
            )

        if self.name in self.response_descriptions:
            tpl = self.response_descriptions[self.name]
            params['response_description'] = tpl.format(
                article=resource_class.name_article,
                name=resource_class.verbose_name,
                pluralname=resource_class.verbose_name_plural
            )

        hints = typing.get_type_hints(resource_class.handle)
        returns = hints.get('return')
        origin = typing.get_origin(returns)
        if origin == list:
            params['response_model'] = returns
        elif inspect.isclass(returns) and issubclass(returns, (pydantic.BaseModel, types.UnionType)):
            params['response_model'] = returns

        return params

    def get_openapi_schema(
        self,
        resource_class: typing.Type[IResource]
    ) -> typing.Dict[str, typing.Any]:
        if self.name not in {"create", "update", "replace"}:
            return {}

        schema: typing.Dict[str, typing.Any] = {}
        models = typing.get_args(resource_class.model)
        oneOf: list[dict[str, typing.Any]] = []
        if models:
            schema['oneOf'] = oneOf
            for model in typing.get_args(models):
                if not inspect.isclass(model)\
                or not issubclass(model, pydantic.BaseModel):
                    raise TypeError(
                        f"{resource_class.__name__}.model must be a subclass of "
                        "pydantic.BaseModel or subclass thereof."
                    )
                oneOf.append(model.schema())
        elif inspect.isclass(resource_class.model)\
        and issubclass(resource_class.model, pydantic.BaseModel):
            schema = resource_class.model.schema()
        else:
            # TODO: Inspect the parameters for pydantic implementations.
            # This should also support plain annotations.
            sig = inspect.signature(self.handler.handle) # type: ignore
            params = sig.parameters
            models: list[type[pydantic.BaseModel]] = []
            for name, value in params.items():
                if name == 'self' or not inspect.isclass(value.annotation):
                    continue
                if not issubclass(value.annotation, pydantic.BaseModel):
                    continue
                models.append(value.annotation)
            if len(models) > 1:
                raise NotImplementedError("Discovery of multiple models is not implemented.")
            if models:
                schema = models[0].schema()
        return {
            'requestBody': {
                'content': {
                    p.media_type: {
                        'schema': schema
                    }
                    for p in resource_class.parsers
                    if schema
                }
            }
        }

    def get_response_description(
        self,
        resource_class: typing.Type[IResource],
        action: str
    ) -> str:
        return self.response_descriptions[self.name].format(
            article=resource_class.name_article,
            name=resource_class.name,
            pluralname=resource_class.pluralname
        )



"""Declares :class:`JWKSEndpoint`."""
import typing

from ckms.types import JSONWebKeySet

from .endpoint import Endpoint
from .params import ServerJWKS


class JWKSEndpoint(Endpoint):
    """Single-tenant endpoint exposing the public keys used by the server to
    verify signatures and encrypt data, as a JSON Web Key Set (JWKS).
    """
    __module__: str = 'cbra.ext.oauth2'
    description: str = (
        "The **JWKS Endpoint** returns a JSON object containing the "
        "public keys that clients may use to encrypt data it sends "
        "to the authorization server, and to verify signatures created "
        "by the authorization server."
    )
    disable_options: bool = False
    methods: set[str] = {'GET'}
    options_description: str = (
        "Communicates the allowed methods and CORS options for "
        "the **JWKS Endpoint**."
    )
    summary: str = 'JSON Web Key Set (JWKS)'
    response_model: type[JSONWebKeySet] = JSONWebKeySet
    response_description: str = "The server public keys."

    def __init__(self, jwks: JSONWebKeySet = ServerJWKS):
        self.jwks = jwks

    async def handle(self) -> dict[str, typing.Any]:
        return self.jwks.dict(exclude_defaults=True)


"""Declares :class:`DefaultPolicy`."""
from typing import Any

from .basecorspolicy import BaseCorsPolicy


class DefaultPolicy(BaseCorsPolicy):
    __module__: str = 'cbra.cors'

    @classmethod
    def new(cls, **attrs: Any) -> type['DefaultPolicy']:
        return type(cls.__name__, (cls,), attrs)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`WebhookEndpoint`."""
import base64
import binascii
import hmac
import warnings
from typing import Any
from typing import NoReturn

from cbra.cors import NullCorsPolicy
from cbra.endpoint import Endpoint
from cbra.exceptions import WebhookMessageRejected
from cbra.types import ICorsPolicy
from cbra.utils import DeferredException
from .models import WebhookResponse
from .webhookendpointmetaclass import WebhookEndpointMetaclass


class WebhookEndpoint(Endpoint, metaclass=WebhookEndpointMetaclass):
    """A :class:`~cbra.Endpoint` implementation that provides a base to
    implement incoming webhook handlers.
    """
    __module__: str = 'cbra.ext.webhook'
    cors_policy: type[ICorsPolicy] = NullCorsPolicy
    method: str = "POST"
    tags: list[str] = ["Webhooks"]
    summary: str = "Webhook events endpoint"
    response_description: str = "The event is accepted for upstream processing."
    response_model: type[WebhookResponse] = WebhookResponse

    @classmethod
    async def create_webhooks(
        cls,
        *,
        hooks: list[dict[str, Any]],
        secret: str,
        **kwargs: Any
    ) -> list[dict[str, Any]]:
        """Take a list of webhook definitions and ensure that they are present
        at the remote system. Return the list of definitions, enriched with data
        returned from the remote system, if any.
        """
        raise NotImplementedError

    def decode_signature(self, signature: str) -> bytes | DeferredException:
        """Decode a Base64-encoded string containing a signature of the request
        payload.
        """
        try:
            return base64.b64decode(signature)
        except (binascii.Error, TypeError, ValueError):
            return WebhookMessageRejected.defer(
                message="The signature is malformed.",
                detail=(
                    "The signature enclosed in the X-WC-Webhook-Signature "
                    "request header could not be interpreted as Base64."
                )
            )

    def get_hmac_secret(self) -> bytes:
        """Return a byte-sequence containing the secret used to verify
        the Hash-based Message Authenticate Code (HMAC) that WooCommerce
        creates in order to secure the request.
        """
        raise NotImplementedError

    async def handle( # type: ignore
        self,
        dto: Any
    ) -> WebhookResponse:
        on_event = getattr(self, 'on_event', self.sink)
        try:
            result = await on_event(dto) or {'status': 'accepted'}
        except Exception as exception:
            self.logger.exception(
                "Caught fatal %s while handling webhook event",
                type(exception).__name__
            )
            result = {'status': 'failed'}
        return WebhookResponse(**result)

    def log_event(self, dto: Any) -> None:
        """Logs the incoming webhook event. This method must be overridden
        as the default implementation raises :exc:`NotImplementedError`.
        """
        raise NotImplementedError

    async def sink(self, dto: Any) -> dict[str, Any]:
        """Invoked for webhook events that do not have any defined
        handlers.
        """
        warnings.warn(f'No handler for event {type(dto).__name__}')
        return {'status': 'dropped'}

    async def verify_request(
        self,
        signature: bytes,
        secret: bytes
    ) -> NoReturn | None:
        """Verifies that a SHA-256 digest of the request body was used to
        create a Hash-based Message Authentication Code (HMAC) using
        `secret`.
        """
        try:
            is_valid = hmac.compare_digest(
                hmac.digest(secret, await self.request.body(), 'sha256'),
                bytes(signature)
            )
        except (TypeError, ValueError): # pragma: no cover
            is_valid = False
        if not is_valid:
            raise WebhookMessageRejected(
                message="The request signature is not valid.",
                detail=(
                    "The signature provided attached to the request "
                    "did not validate against a HMAC-SHA256 comparison "
                    "using the pre-shared secret."
                )
            )


"""Declares :class:`ServerTokenIssuer`."""
from typing import cast
from typing import Any

from ckms.jose import ClaimSet
from ckms.jose import PayloadCodec

from .types import RFC9068Token


class ServerTokenIssuer:
    """Provides an interface to sign and decrypt tokens using the
    authorization servers' configured default keys and algorithms.
    """
    __module__: str = 'cbra.ext.oauth2'
    algorithm: str
    codec: PayloadCodec
    default_ttl: int = 60
    using: str

    async def issue_access_token(
        self,
        claims: dict[str, Any] | RFC9068Token,
        header: dict[str, Any] | None = None,
        ttl: int | None = None,
        model: type[ClaimSet] = RFC9068Token
    ) -> str:
        """Sign an :rfc:`9068` access token using the default signing
        key configured for the authorization server.
        """
        ttl = ttl or self.default_ttl
        if not isinstance(claims, (RFC9068Token, model)):
            claims = model.strict(ttl=ttl, **claims)
        header = header or {}
        header.setdefault('typ', "at+jwt")
        payload = self.codec.encode(claims)
        jws = await payload.sign(
            algorithm=self.algorithm,
            using=self.using,
            header=header
        )
        return bytes.decode(bytes(jws), 'ascii')


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from .clientkeysendpoint import ClientKeysEndpoint
from .personaldatahandler import PersonalDataHandler
from .resource import Resource
from .service import Service
from .serviceclient import ServiceClient


__all__: list[str] = [
    'ClientKeysEndpoint',
    'Resource',
    'PersonalDataHandler',
    'PersonalDataService',
    'Service',
    'ServiceClient'
]


CurrentService: ServiceClient = fastapi.Depends(ServiceClient.current)

PersonalDataService: PersonalDataHandler = fastapi.Depends()


# pylint: skip-file
# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
import pytest
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken
from httpx import AsyncClient

from cbra import Application
from examples.google import ExampleGoogleEndpoint
from .conftest import get_id_token


@pytest.mark.asyncio
async def test_authenticate_with_token(
    id_token: str,
    client: AsyncClient
):
    _, claims = await PayloadCodec().jwt(id_token)
    response = await client.post( # type: ignore
        url='/',
        headers={'Authorization': f'Bearer {id_token}'}
    )
    dto = response.json()
    assert isinstance(claims, JSONWebToken)
    assert response.status_code == 200
    assert dto.get('email') == claims.extra.get('email')


@pytest.mark.asyncio
async def test_bearer_token_is_required(
    client: AsyncClient
):
    response = await client.post( # type: ignore
        url='/'
    )
    dto = response.json()
    assert response.status_code == 401
    assert dto['spec']['code'] == 'AUTHENTICATION_REQUIRED'


@pytest.mark.asyncio
async def test_invalid_audience_is_rejected(
    client: AsyncClient
):
    id_token = get_id_token("https://example.com")
    response = await client.post( # type: ignore
        url='/',
        headers={'Authorization': f'Bearer {id_token}'}
    )
    dto = response.json()
    assert response.status_code == 403
    assert dto['spec']['code'] == "WRONG_AUDIENCE"


@pytest.mark.asyncio
async def test_invalid_scheme_is_rejected(
    id_token: str,
    client: AsyncClient
):
    response = await client.post( # type: ignore
        url='/',
        headers={'Authorization': f'Basic {id_token}'}
    )
    dto = response.json()
    assert response.status_code == 403
    assert dto['spec']['code'] == "INVALID_AUTHORIZATION_SCHEME"


@pytest.mark.asyncio
async def test_no_whitelisted_service_accounts_returns_403(
    app: Application,
    id_token: str,
    client: AsyncClient
):
    Endpoint = ExampleGoogleEndpoint.new( # type: ignore
        service_accounts=set()
    )
    app.add(Endpoint, base_path='/whitelisted')
    response = await client.post( # type: ignore
        url='/whitelisted',
        headers={'Authorization': f'Bearer {id_token}'}
    )
    dto = response.json()
    assert response.status_code == 403
    assert dto['spec']['code'] == "FORBIDDEN"


@pytest.mark.asyncio
async def test_different_whitelisted_service_accounts_returns_403(
    app: Application,
    id_token: str,
    client: AsyncClient
):
    Endpoint = ExampleGoogleEndpoint.new( # type: ignore
        service_accounts={"foo@bar.baz"}
    )
    app.add(Endpoint, base_path='/whitelisted')
    response = await client.post( # type: ignore
        url='/whitelisted',
        headers={'Authorization': f'Bearer {id_token}'}
    )
    dto = response.json()
    assert response.status_code == 403
    assert dto['spec']['code'] == "FORBIDDEN"


@pytest.mark.asyncio
async def test_allow_unauthenticated(
    app: Application,
    id_token: str,
    client: AsyncClient
):
    Endpoint = ExampleGoogleEndpoint.new( # type: ignore
        require_authentication=False
    )
    app.add(Endpoint, base_path='/no-authentication')
    response = await client.post( # type: ignore
        url='/no-authentication',
    )
    dto = response.json()
    assert response.status_code == 200
    assert dto.get('email') == "none"


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ckms.core import Keychain
from ckms.types import JSONWebKeySet

import cbra
from cbra.cors import AnonymousReadCorsPolicy
from cbra.cors import CorsPolicyType


class ClientKeysEndpoint(cbra.Endpoint):
    """Single-tenant endpoint exposing the public keys used with the OAuth 2.x
    client of the server.
    """
    __module__: str = 'cbra.ext.service'
    cors_policy: type[CorsPolicyType] = AnonymousReadCorsPolicy
    name: str = 'service.oauth2-jwks'
    description: str = (
        "Provides the JSON Web Key Set (JWKS) that the server used to "
        "assert the ownership of its OAuth 2.x client."
    )
    keychain: Keychain
    method: str = 'GET'
    mount_path: str = '.well-known/oauth-client-jwks'
    summary: str = 'OAuth 2.x Client JWKS'
    response_model: type[JSONWebKeySet] = JSONWebKeySet
    response_model_by_alias: bool = True
    response_model_exclude_none: bool = True
    response_description: str = "The OAuth 2.x client public keys."
    tags: list[str] = ["Server metadata"]
    with_options: bool = True

    async def handle(self) -> JSONWebKeySet:
        return self.keychain.as_jwks(private=False)


"""Declares :class:`FormParser`."""
import urllib.parse

import fastapi

from cbra.types import IParser


class FormParser(IParser):
    """
    Parser for form data.
    """
    __module__: str = 'cbra.parsers'
    media_type: str = 'application/x-www-form-urlencoded'
    charset: str = "utf-8"

    async def parse(self,
        request: fastapi.Request,
        media_type: str = None,
        parser_context: dict = None
    ) -> dict:
        """Parses the incoming bytestream as a URL encoded form, and returns
        the resulting dictionary.
        """
        parser_context = parser_context or {}
        encoding = parser_context.get('encoding', self.charset)
        body = bytes.decode(await request.body(), encoding)
        return {k: v for k, v in urllib.parse.parse_qsl((body or ''), True)}



"""Declares :class:`ResourceOptions`."""
from inspect import Parameter
from typing import Any

from cbra.exceptions import get_exception_headers
from cbra.options import Options


class ResourceOptions(Options):
    is_detail: bool
    path_parameters: list[Parameter] = []

    @classmethod
    def get_path_signature(cls, detail: bool = False) -> list[Parameter]:
        return cls.path_parameters

    @classmethod
    def get_responses(cls) -> dict[int, Any]:
        responses = super().get_responses()
        if cls.is_detail:
            responses[404] = {
                'headers': get_exception_headers(404),
                'description': (
                    "The resource specified by the path parameter(s) does not exist."
                )
            }
        return responses


"""Declares :class:`IUserAgent`."""
import typing

import fastapi


class IUserAgent:
    __module__: str = 'cbra.types'

    def __init__(
        self,
        user_agent: typing.Optional[str] = fastapi.Header(
            default=None,
            title="User agent",
            alias="User-Agent",
            description=(
                "Identifies the application, operating system, vendor, "
                "and/or version of the requesting user agent."
            )
        )
    ):
        self.user_agent = user_agent



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import decimal
import uuid
from typing import Any

import json

class JSONEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time/timedelta,
    decimal types, generators and other basic python objects.
    """
    def default(self, o: Any) -> str | tuple[Any]:
        if isinstance(o, datetime.datetime):
            representation = o.isoformat()
            if representation.endswith('+00:00'):
                representation = representation[:-6] + 'Z'
            return representation
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif isinstance(o, datetime.time):
            if o.utcoffset() is not None:
                raise ValueError("JSON can't represent timezone-aware times.")
            representation = o.isoformat()
            return representation
        elif isinstance(o, datetime.timedelta):
            return str(o.total_seconds())
        elif isinstance(o, decimal.Decimal):
            return str(o)
        elif isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, bytes):
            return o.decode()
        elif hasattr(o, 'tolist'):
            # Numpy arrays and array scalars.
            return o.tolist()
        elif hasattr(o, '__iter__'):
            return tuple(item for item in o)
        return super().default(o)


"""Declares :class:`ClientSubject`."""
from typing import Any

from .types import ISubject


class ClientSubject(ISubject):
    """A subject implementation that represents a client."""
    __module__: str = 'cbra.ext.oauth2'

    def __init__(self, client_id: str, **kwargs: Any):
        self.sub = client_id

    def get_identifier(self, public: bool = False) -> int:
        return self.sub


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
import base64
import json
from typing import Any

import pytest
from httpx import AsyncClient


class IMessageTest:
    event_type: str
    resource_type: str
    signature_header: str

    def get_request_params(self) -> dict[str, Any]:
        raise NotImplementedError

    def test_content_is_json_parseable(self) -> None:
        params = self.get_request_params()
        json.loads(bytes.decode(params['content']))

    @pytest.mark.asyncio
    async def test_accept_event(
        self,
        client: AsyncClient
    ) -> None:
        raise NotImplementedError

    @pytest.mark.asyncio
    async def test_disallowed_event_is_rejected(
        self,
        client: AsyncClient
    ) -> None:
        pytest.skip()

    @pytest.mark.asyncio
    async def test_missing_signature(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        self.remove_signature_header(params)
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 403, response.text
        assert dto['spec']['code'] == 'WEBHOOK_EVENT_REJECTED'
        assert dto['spec']['message'] == "The request signature is not valid."

    @pytest.mark.asyncio
    async def test_malformed_signature(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        self.add_malformed_signature(params)
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 403, response.text
        assert dto['spec']['code'] == 'WEBHOOK_EVENT_REJECTED'
        assert dto['spec']['message'] == "The signature is malformed."

    @pytest.mark.asyncio
    async def test_mismatching_signature(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        self.add_mismatching_signature(params)
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 403, response.text
        assert dto['spec']['code'] == 'WEBHOOK_EVENT_REJECTED'
        assert dto['spec']['message'] == "The request signature is not valid."

    def add_malformed_signature(self, params: dict[str, Any]) -> dict[str, Any]:
        params['headers'][self.signature_header] = 'foo'

    def add_mismatching_signature(self, params: dict[str, Any]) -> dict[str, Any]:
        params['headers'][self.signature_header] = base64.b64encode(b'foo')

    def remove_signature_header(self, params: dict[str, Any]) -> dict[str, Any]:
        del params['headers'][self.signature_header]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`ICorsPolicy`."""
import typing
from typing import Any

import fastapi
import fastapi.params

from cbra.headers import ACCESS_CONTROL_REQUEST_HEADERS
from cbra.headers import ACCESS_CONTROL_REQUEST_METHOD
from cbra.headers import ORIGIN
from cbra.params import CurrentServer


class ICorsPolicy:
    __module__: str = 'cbra.types'
    allow_credentials: bool = False
    allowed_methods: set[str] = set()
    allowed_headers: set[str] = set()
    allowed_response_headers: set[str] = set()
    allowed_origins: set[str] = set()
    max_age: int = 86400
    origin: str | None
    origin_header: fastapi.params.Depends | fastapi.params.Header = ORIGIN # type: ignore
    request: fastapi.Request
    request_headers: set[str]
    request_methods: set[str]
    server: str
    OriginHeader: str | None = ORIGIN

    @classmethod
    def as_options(
        cls,
        request: fastapi.Request,
        server: str = CurrentServer,
        origin: str | None = ORIGIN,
        request_headers: typing.Optional[str] = ACCESS_CONTROL_REQUEST_HEADERS,
        request_methods: typing.Optional[str] = ACCESS_CONTROL_REQUEST_METHOD
    ) -> 'ICorsPolicy':
        instance = cls(
            request=request,
            server=server,
            origin=origin,
        )
        instance.request_headers = set(
            filter(bool, map(str.strip, str.split(request_headers or '', ',')))
        )
        instance.request_methods = set(
            filter(bool, map(str.strip, str.split(request_methods or '', ',')))
        )
        return instance

    @classmethod
    def configure(
        cls,
        *,
        allowed_headers: set[str] = set(),
        allowed_methods: set[str] = set()
    ) -> type['ICorsPolicy']:
        allowed_methods.add("OPTIONS")
        return type(cls.__name__, (cls,), {
            '__module__': cls.__module__,
            'allowed_headers': allowed_headers,
            'allowed_methods': allowed_methods
        })

    @classmethod
    def get_response_headers(cls) -> dict[str, Any]:
        return {}

    def __init__(
        self,
        request: fastapi.Request,
        server: str = CurrentServer,
        origin: str | None = ORIGIN
    ):
        self.origin = origin
        self.request = request
        self.server = server

    async def add_response_headers(
        self,
        request: fastapi.Request,
        response: fastapi.Response,
        handler: typing.Any,
    ) -> None:
        raise NotImplementedError

    async def check_cors_allowed(
        self,
        handler: typing.Any,
        origin: typing.Optional[str]
    ) -> typing.Optional[typing.NoReturn]:
        raise NotImplementedError

    async def get_allowed_origins(self) -> set[str]:
        """Return the set of allowed origins."""
        raise NotImplementedError

    async def process_request(
        self,
        request: fastapi.Request,
        handler: typing.Any,
        origin: typing.Optional[str] = None
    ) -> typing.Union[bool, typing.NoReturn]:
        raise NotImplementedError

    async def process_response(
        self,
        request: fastapi.Request,
        response: fastapi.Response,
        handler: typing.Any,
        origin: typing.Optional[str] = None
    ) -> typing.Optional[typing.NoReturn]:
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import aorta.transport
from unimatrix.conf import settings # type: ignore

from .models import DependencyList
from .provider import _default # type: ignore


DEPENDENCIES: list[Any] = getattr(
    settings, 'DEPENDENCIES', [] # type: ignore
)


async def init():
    """Loads the preconfigured dependencies in the default provider."""
    await _default.load_many(DependencyList(items=DEPENDENCIES))
    if not _default.is_satisfied('MessageTransport'):
        _default.provide('MessageTransport', aorta.transport.NullTransport())



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`ProductEvent`."""
import pydantic

from .event import Event
from .productstatustype import ProductStatusType
from .producttype import ProductType


class ProductEvent(Event):
    type: ProductType = pydantic.Field(
        default=...,
        title="Product type",
        description=(
            "Describes the type of product on which an event occurred."
        )
    )

    status: ProductStatusType = pydantic.Field(
        default=...,
        title="Status",
        description=(
            "The current status of the product resource."
        )
    )

    class Config:
        title: str = "WooCommerceProductEvent"


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from ckms.jose import PayloadCodec

from cbra.exceptions import Forbidden
from .params import ClientRepository
from .params import ServerCodec
from .types import ClientAssertion
from .types import IClient
from .types import IClientRepository


__all__: list[str] = ['CurrentClient']


async def get_current_client(
    request: fastapi.Request,
    assertion: ClientAssertion,
    clients: IClientRepository = ClientRepository,
    codec: PayloadCodec = ServerCodec
) -> IClient:
    if assertion.client_assertion is None:
        raise Forbidden
    jws, jwt = await codec.jwt(assertion.client_assertion)
    if jwt.iss is None:
        raise Forbidden
    client = await clients.get(jwt.iss)
    if client is None:
        raise Forbidden

    # Verify that the JWT has the proper audience and was signed by
    # this client.
    jwt.verify(audience={str(request.url)})
    if not await client.verify_jws(jws):
        raise Forbidden(
            message="The signature is not valid for the asserted client."
        )
    return client


CurrentClient: IClient = fastapi.Depends(get_current_client)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .authorizationexception import AuthorizationException


class PolicyFailure(AuthorizationException):
    __module__: str = 'cbra.ext.oauth2.types'
    name: str
    allow: list[Any]

    def __init__(self, *, name: str, allow: list[Any] | None = None):
        self.name = name
        self.allow = allow or []



"""Declares :class:`BaseSubject`."""
from collections import defaultdict

from cbra.utils import docstring
from .isubject import ISubject


class BaseSubject(ISubject):
    __module__: str = 'cbra.ext.oauth2.types'
    sub: int | str
    client_id: str | None = None
    scope: set[str] = set()
    client_scope: defaultdict[str | None, set[str]] = defaultdict(set)

    @docstring(ISubject)
    def allows_scope(self, scope: set[str]) -> bool:
        return scope <= self.client_scope[self.client_id]